#include "descriptors.h"
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/highgui/highgui.hpp"
#include <stdio.h>
#include <unistd.h>
#include <cv.h>
#include <fstream>
#include <iostream>

using namespace cv;
using namespace std;

int descriptorCN( char *imageFile, char *detectorFile, int patch_size, int scalef, double sigma, int type, std::ostream &outFile )
{
	int m,k,n,p,index_px,blue,green,red,cont;
	int cx,cy,x,y,r,side,w,h,xi,yi,imgM,imgN;
	int outCols,border,hue_bin;
	double d,scalex,scaley,px,py,dx,dy,weight,hue,sat,lumEnergy,sum;
	int detRows,detCols,detCols2,w2cRows,w2cCols;
	Mat image,superImage,det,w2c,out,gaussian,superMask,smooth;
	string s;
	ifstream fdin;
	Vec3b pixel;
	
	// Read Image File
	image = imread(imageFile, 1);
	if(image.empty())
	{
		printf("Cannot read image file: %s\n", imageFile);
		return -1;
	}
	imgM = image.rows;
	imgN = image.cols;
	
	// Read Detector File
	fdin.open(detectorFile);
	if(fdin.fail())
	{
		printf("Cannot read detector file: %s\n", detectorFile);
		return -1;
	}
	// Check header
	fdin >> s;
	fdin >> detCols2;
	fdin >> detRows;
	if(detRows < 1)
	{
		printf("No rows specified in file: %s\n", detectorFile);
		return -1;
	}
	detCols = 5; 
	det.create(detRows,detCols,CV_64F);
	for(m=0; m<detRows; m++)
	{
		// "<CIRCLE x y scale orientation cornerness>;;"
		fdin.ignore(100,' ');
		for(k=0 ; k<detCols ; k++)
		{
			fdin >> d;
			det.at<double>(m,k) = d;
		}
		fdin.ignore(10000,';');
		fdin.ignore(10000,';');
	}
	fdin.close();
	
	// Type 0: Color Naming
	string w2cFile = "w2c/w2c.txt";
	if (type==1)
	{
		// Type 1: JOSA Color Naming
		w2cFile = "w2c/LUT_JOSA.txt";
	}
	// Read Lookup Table
	fdin.open(w2cFile.c_str());
	if(fdin.fail())
	{
		printf("Cannot read lookup table file: %s\n",w2cFile.c_str());
		return -1;
	}
	fdin >> w2cRows;
	fdin >> w2cCols;
	w2c.create(w2cRows,w2cCols,CV_64F);
	for(m=0; m<w2cRows; m++)
	{
		for(k=0; k<w2cCols; k++)
		{
			fdin >> d;
			w2c.at<double>(m,k) = d;
		}
	}
	fdin.close();
	
	// Output number of columns depending on descriptor
	outCols = detCols+w2cCols-3;
	
	// Detectors out of image boundaries
	border = 0;
	for(m=0; m<detRows; m++)
	{
		r = round(det.at<double>(m,2)*scalef);
		if(r>border)
		{
			border = r;
		}
	}
	border = 3*border;
	
	// Create Super Image for treating detectors out of image bounds
	superImage.create(imgM+2*border,imgN+2*border,CV_8UC3);
	superMask.create(imgM+2*border,imgN+2*border,CV_64F);
	for(m=0; m<border; m++)
	{
		for(k=0; k<border; k++)
		{
			// top-left
			superImage.at<Vec3b>(m,k) = image.at<Vec3b>(0,0);
			superMask.at<double>(m,k) = 0.;
			// bottom-left
			superImage.at<Vec3b>(m+border+imgM,k) = image.at<Vec3b>(imgM-1,0);
			superMask.at<double>(m+border+imgM,k) = 0.;
			// top-right
			superImage.at<Vec3b>(m,k+border+imgN) = image.at<Vec3b>(0,imgN-1);
			superMask.at<double>(m,k+border+imgN) = 0.;
			// bottom-right
			superImage.at<Vec3b>(m+border+imgM,k+border+imgN) = image.at<Vec3b>(imgM-1,imgN-1);
			superMask.at<double>(m+border+imgM,k+border+imgN) = 0.;
		}
	}
	for(m=0; m<imgM; m++)
	{
		for(k=0; k<border; k++)
		{
			// middle-left
			superImage.at<Vec3b>(m+border,k) = image.at<Vec3b>(m,0);
			superMask.at<double>(m+border,k) = 0.;
			// middle-right
			superImage.at<Vec3b>(m+border,k+border+imgN) = image.at<Vec3b>(m,imgN-1);
			superMask.at<double>(m+border,k+border+imgN) = 0.;
		}
	}
	for(m=0; m<border; m++)
	{
		for(k=0; k<imgN; k++)
		{
			// top-center
			superImage.at<Vec3b>(m,k+border) = image.at<Vec3b>(0,k);
			superMask.at<double>(m,k+border) = 0.;
			// bottom-center
			superImage.at<Vec3b>(m+border+imgM,k+border) = image.at<Vec3b>(imgM-1,k);
			superMask.at<double>(m+border+imgM,k+border) = 0.;
		}
	}
	for(m=0; m<imgM; m++)
	{
		for(k=0; k<imgN; k++)
		{
			// middle-center
			superImage.at<Vec3b>(m+border,k+border) = image.at<Vec3b>(m,k);
			superMask.at<double>(m+border,k+border) = 1.;
		}
	}
	
	image.release();
	
	// Gaussian
	gaussian.create(patch_size,patch_size,CV_64F);
	for(m=0; m<gaussian.rows; m++)
	{
		for(k=0; k<gaussian.cols; k++)
		{
			if(sigma == 0)
			{
				gaussian.at<double>(m,k) = 1.;
			}
			else
			{
				gaussian.at<double>(m,k) = exp((-(pow(m - ((patch_size - 1)/2), 2) + pow(k - ((patch_size - 1)/2), 2)))/(2*sigma*sigma));
			}
		}
	}
	
	// Output matrix
	Mat imoutR, imoutG, imoutB, patchMask;
	imoutR.create(patch_size,patch_size,CV_64F);
	imoutG.create(patch_size,patch_size,CV_64F);
	imoutB.create(patch_size,patch_size,CV_64F);
	patchMask.create(patch_size,patch_size,CV_64F);
	
	out.create(detRows,outCols,CV_64F);
	
	// Apply to every ROI
	for(m=0; m<out.rows; m++)
	{
		// Initialize output row
		for(k=0; k<detCols; k++)
		{
			out.at<double>(m,k) = det.at<double>(m,k);
		}
		for(k=detCols; k<outCols; k++)
		{
			out.at<double>(m,k) = 0.0;
		}
		// The -1 is because the descriptor comes from MatLab and starts at position 1 instead of 0
		cx = det.at<double>(m,0) + border -1;
		cy = det.at<double>(m,1) + border -1;
		r = round(det.at<double>(m,2)*scalef);
		x = cx - r;
		y = cy - r;
		side = (int) 2*r+1;
		scalex  = (double) side / patch_size;
		scaley  = (double) side / patch_size;
		
		// Check if the roi is not totally inside the image
		if((x+side>superImage.cols) || (y+side>superImage.rows) || (x<0) || (y<0))
		{
			printf("Completely outside of the image detection\n");
			continue;
		}
		
		if(side==patch_size)
		{
			for (k=y; k<y+side; k++)
			{
				for (n=x; n<x+side; n++)
				{
					pixel = superImage.at<Vec3b>(k,n);
					imoutR.at<double>(k-y,n-x) = pixel.val[2];
					imoutG.at<double>(k-y,n-x) = pixel.val[1];
					imoutB.at<double>(k-y,n-x) = pixel.val[0];
					patchMask.at<double>(k-y,n-x) = superMask.at<double>(k,n);
				}
			}
		}
		else
		{
			// Separate the image to resize into 3 channels
			Mat imgR_blk, imgG_blk, imgB_blk, mask;
			imgR_blk.create(side,side,CV_64F);
			imgG_blk.create(side,side,CV_64F);
			imgB_blk.create(side,side,CV_64F);
			mask.create(side,side,CV_64F);
			
			for (k=y; k<y+side; k++)
			{
				for (n=x; n<x+side; n++)
				{
					pixel = superImage.at<Vec3b>(k,n);
					imgR_blk.at<double>(k-y,n-x) = pixel.val[2];
					imgG_blk.at<double>(k-y,n-x) = pixel.val[1];
					imgB_blk.at<double>(k-y,n-x) = pixel.val[0];
					mask.at<double>(k-y,n-x) = superMask.at<double>(k,n);
				}
			}
			
			w = side-1;
			h = side-1;
			
			// Resize roi to patch
			for(k=0; k<patch_size; k++)
			{
				py = k*scaley;
				yi = (int) py;
				dy = py-yi;
				for(n=0 ; n<patch_size; n++)
				{
					px = n*scalex;
					xi = (int) px;
					dx = px-xi;
					if( (xi<w) && (yi<h) )
					{
						imoutR.at<double>(k,n) = (1.-dy) * ((1.-dx)*imgR_blk.at<double>(yi,xi)+dx*imgR_blk.at<double>(yi,xi+1)) + dy * ((1.-dx)*imgR_blk.at<double>(yi+1,xi)+dx*imgR_blk.at<double>(yi+1,xi+1));
						imoutG.at<double>(k,n) = (1.-dy) * ((1.-dx)*imgG_blk.at<double>(yi,xi)+dx*imgG_blk.at<double>(yi,xi+1)) + dy * ((1.-dx)*imgG_blk.at<double>(yi+1,xi)+dx*imgG_blk.at<double>(yi+1,xi+1));
						imoutB.at<double>(k,n) = (1.-dy) * ((1.-dx)*imgB_blk.at<double>(yi,xi)+dx*imgB_blk.at<double>(yi,xi+1)) + dy * ((1.-dx)*imgB_blk.at<double>(yi+1,xi)+dx*imgB_blk.at<double>(yi+1,xi+1));
						patchMask.at<double>(k,n) = (1.-dy) * ((1.-dx)*mask.at<double>(yi,xi)+dx*mask.at<double>(yi,xi+1)) + dy * ((1.-dx)*mask.at<double>(yi+1,xi)+dx*mask.at<double>(yi+1,xi+1));
					}
					else if( (xi<=w) && (yi<h) )
					{
						imoutR.at<double>(k,n) = (1.-dy) * ((1.-dx)*imgR_blk.at<double>(yi,w) +dx*imgR_blk.at<double>(yi,w))    + dy * ((1.-dx)*imgR_blk.at<double>(yi+1,w) +dx*imgR_blk.at<double>(yi+1,w));
						imoutG.at<double>(k,n) = (1.-dy) * ((1.-dx)*imgG_blk.at<double>(yi,w) +dx*imgG_blk.at<double>(yi,w))    + dy * ((1.-dx)*imgG_blk.at<double>(yi+1,w) +dx*imgG_blk.at<double>(yi+1,w));
						imoutB.at<double>(k,n) = (1.-dy) * ((1.-dx)*imgB_blk.at<double>(yi,w) +dx*imgB_blk.at<double>(yi,w))    + dy * ((1.-dx)*imgB_blk.at<double>(yi+1,w) +dx*imgB_blk.at<double>(yi+1,w));
						patchMask.at<double>(k,n) = (1.-dy) * ((1.-dx)*mask.at<double>(yi,w) +dx*mask.at<double>(yi,w))    + dy * ((1.-dx)*mask.at<double>(yi+1,w) +dx*mask.at<double>(yi+1,w));
					}
					else if( (xi<w) && (yi<=h) )
					{
						imoutR.at<double>(k,n) = (1.-dy) * ((1.-dx)*imgR_blk.at<double>(yi,xi)+dx*imgR_blk.at<double>(yi,xi+1)) + dy * ((1.-dx)*imgR_blk.at<double>(h,xi)+dx*imgR_blk.at<double>(h,xi+1));
						imoutG.at<double>(k,n) = (1.-dy) * ((1.-dx)*imgG_blk.at<double>(yi,xi)+dx*imgG_blk.at<double>(yi,xi+1)) + dy * ((1.-dx)*imgG_blk.at<double>(h,xi)+dx*imgG_blk.at<double>(h,xi+1));
						imoutB.at<double>(k,n) = (1.-dy) * ((1.-dx)*imgB_blk.at<double>(yi,xi)+dx*imgB_blk.at<double>(yi,xi+1)) + dy * ((1.-dx)*imgB_blk.at<double>(h,xi)+dx*imgB_blk.at<double>(h,xi+1));
						patchMask.at<double>(k,n) = (1.-dy) * ((1.-dx)*mask.at<double>(yi,xi)+dx*mask.at<double>(yi,xi+1)) + dy * ((1.-dx)*mask.at<double>(h,xi)+dx*mask.at<double>(h,xi+1));
					}
					else if( (xi<=w) && (yi<=h) )
					{
						imoutR.at<double>(k,n) = (1.-dy) * ((1.-dx)*imgR_blk.at<double>(yi,w) +dx*imgR_blk.at<double>(yi,w))    + dy * ((1.-dx)*imgR_blk.at<double>(h,w) +dx*imgR_blk.at<double>(h,w));
						imoutG.at<double>(k,n) = (1.-dy) * ((1.-dx)*imgG_blk.at<double>(yi,w) +dx*imgG_blk.at<double>(yi,w))    + dy * ((1.-dx)*imgG_blk.at<double>(h,w) +dx*imgG_blk.at<double>(h,w));
						imoutB.at<double>(k,n) = (1.-dy) * ((1.-dx)*imgB_blk.at<double>(yi,w) +dx*imgB_blk.at<double>(yi,w))    + dy * ((1.-dx)*imgB_blk.at<double>(h,w) +dx*imgB_blk.at<double>(h,w));
						patchMask.at<double>(k,n) = (1.-dy) * ((1.-dx)*mask.at<double>(yi,w) +dx*mask.at<double>(yi,w))    + dy * ((1.-dx)*mask.at<double>(h,w) +dx*mask.at<double>(h,w));
					}
					
					if(patchMask.at<double>(k,n)<0.5)
					{
						patchMask.at<double>(k,n) = 0.;
					}
					else
					{
						patchMask.at<double>(k,n) = 1.;
					}
				}
			}
			
			mask.release();
			imgB_blk.release();
			imgG_blk.release();
			imgR_blk.release();
		}
		
		// Apply CN to all pixels
		cont=0;
		sum=0;
		for(k=0; k<patch_size; k++)
		{
			for(n=0; n<patch_size; n++)
			{
				if(patchMask.at<double>(k,n)==1)
				{
					weight = gaussian.at<double>(k,n);
					blue = imoutB.at<double>(k,n);
					green = imoutG.at<double>(k,n);
					red = imoutR.at<double>(k,n);
					index_px = floor(red/8)+32*floor(green/8)+32*32*floor(blue/8);
					// Add CN probs
					for(p=detCols; p<out.cols; p++)
					{
						out.at<double>(m,p) = out.at<double>(m,p) + weight * w2c.at<double>(index_px,p-2);
					}
					cont++;
					sum = sum + weight;
				}
			}
		}
		// Normalize gaussian
		if(cont>0)
		{
			for(k=detCols; k<outCols; k++)
			{
				out.at<double>(m,k) = out.at<double>(m,k)/sum;
				
			}
		}
	}
	
	// Print output
	// Order of color probabilities: black,blue,brown,grey,green,orange,pink,purple,red,white,yellow
	outFile << "MARC1" << endl;
	outFile << (out.cols - detCols) << endl;
	outFile << out.rows << endl;
	for(m=0; m<out.rows; m++)
	{
		outFile << "<CIRCLE ";
		for(k=0; k<detCols-1; k++)
		{
			outFile << out.at<double>(m,k) << " ";
		}
		outFile << out.at<double>(m,detCols-1) << ">; ";
		for(k=detCols; k<out.cols-1; k++)
		{
			outFile << out.at<double>(m,k) << " ";
		}
		outFile << out.at<double>(m,out.cols-1);
		outFile << ";" << endl;
	}
	
	smooth.release();
	patchMask.release();
	imoutR.release();
	imoutG.release();
	imoutB.release();
	superMask.release();
	gaussian.release();
	out.release();
	w2c.release();
	det.release();
	superImage.release();
	return 0;
}
